module image_transforms

  ! use pnm_class
  use polysample
  use precision
  use rotmatrix_class
  use fundamental_constants, only: radian, pi
  use variable_array, only: push, splice

  implicit none

  private

  character(len=*), parameter :: version = "$Id: image_transforms.f90,v 1.4 2005/07/01 03:22:50 aidan Exp aidan $"

  !! Routines for rotating and/or scaling images. Currently supports
  !! rotating double, real and integer images. There are subroutine and
  !! functional interfaces. The double precision versions accept only
  !! double precision arguments. The real wrappers will accept double
  !! and single precision angles, but the type of the missing value must
  !! match that of the data that is being manipulated. The integer wrapper 
  !! script will accept either, but if two such arguments are supplied they 
  !! must be the same kind. 
  !! 
  !! The code dealing with direct support for pnm objects has been commented
  !! out (it is probably not in working order either). Rotating pnm's is
  !! problematic -- what if we have a ppm? Really we should rotate all three
  !! fields simultaneously to save computing the complicated overlaps for each
  !! channel. We do not do this currently. More work is required. Possibly
  !! this rotation operation should be moved into the pnm module -- then it
  !! could operate directly on the internal integer data and we wouldn't have
  !! to have a wrapper for it in this module. Hmmmmmm ....

  !! $Log: image_transforms.f90,v $
  !! Revision 1.4  2005/07/01 03:22:50  aidan
  !! Added rotational averaging routines, both subroutine and functional interfaces.
  !!
  !! Revision 1.3  2005/06/29 03:25:03  aidan
  !! Forgot to delete the set_coverage, get_coverage routines from the interface.
  !! Fixed.
  !!
  !! Revision 1.2  2005/06/29 03:20:39  aidan
  !! Added support for specifying minimum coverage.
  !!
  !! Revision 1.1  2004/11/17 02:34:33  aidan
  !! Initial revision
  !!
  !!

  ! Subroutine interface
  interface rotate_image
     module procedure rotate_image_real8_sub, rotate_inplace, &
          rotate_inplace_integer_wrap, rotate_inplace_integerd_wrap, &
          rotate_inplace_real_wrap, rotate_inplace_reald_wrap, &
          rotate_image_integer_wrap, rotate_image_integerd_wrap, &
          rotate_image_real_wrap, rotate_image_reald_wrap
  end interface

  interface rotave_image
     module procedure rotave_real8_sub, rotave_resize_real8_sub
     module procedure rotave_real8_sub_nomiss, rotave_resize_real8_sub_nomiss
     module procedure rotave_integer_sub, rotave_real_sub
     module procedure rotave_real_inplace_sub, rotave_integer_inplace_sub
  end interface

  ! Functional interface
  interface rotimage
     module procedure rotate_resize_image_real8_fn, rotate_image_real8_fn, &
          rotate_resize_image_real_fn, rotate_image_real_fn, &
          rotate_resize_image_reald_fn, rotate_image_reald_fn, &
          rotate_resize_image_integer_fn, rotate_image_integer_fn, &
          rotate_resize_image_integerd_fn, rotate_image_integerd_fn
  end interface

  interface rotave
     module procedure rotave_real8_fn, rotave_real_fn, rotave_integer_fn
  end interface

  interface centre
     module procedure centre_real8_inplace
  end interface

  public :: rotate_image, rotimage, rotave_image, rotave, centre

  logical, parameter :: debug =  .FALSE. !.TRUE.

contains

  ! This is the main workhorse routine for this module. Most of the other
  ! routines are wrappers for this routine.

  subroutine rotate_image_real8_sub(data, angle, output, missingvalue, coverage)

    ! Interface variables
    real(kind=rd_kind), dimension(:,:), intent(in)  :: data
    real(kind=rd_kind), intent(in)                  :: angle
    real(kind=rd_kind), dimension(:,:), intent(out) :: output
    real(kind=rd_kind), intent(in), optional        :: missingvalue
    real(kind=rs_kind), intent(in), optional        :: coverage

    ! Local variables
    real(kind=rd_kind), dimension(3,(size(output,1)+1)*(size(output,2)+1)) :: vertices
    integer :: i, nx, ny, newnx, newny

    type (rotmatrix) :: R

    if (debug) print *,'in rotate_image'

    nx = size(data,1)
    ny = size(data,2)
    newnx = size(output,1)
    newny = size(output,2)

    if (debug) print *,nx,ny,newnx,newny

    ! Reproduce the current vertices (runs from 0 -> nx, 0 -> ny)
    vertices(1,:) = pack(spread(real((/ (i, i = 0, newnx) /),rd_kind) * &
         (real(nx,rd_kind)/real(newnx,rd_kind)), 2, newny+1),.TRUE.)
    vertices(2,:) = pack(spread(real((/ (i, i = 0, newny) /),rd_kind) * &
         (real(ny,rd_kind)/real(newny,rd_kind)), 1, newnx+1),.TRUE.)
    vertices(3,:) = 0.d0

    if (debug) print *,'Original vertices'
    if (debug) print '(2F)',vertices(1:2,1:10)

    ! stop

    vertices(1,:) = vertices(1,:) - real(nx,rd_kind)/2.d0
    vertices(2,:) = vertices(2,:) - real(ny,rd_kind)/2.d0
    
    if (debug) print *,'Change origin dx dy = ',-nx/2,-ny/2
    if (debug) print '(2F)',vertices(1:2,:)

    ! Make a rotation matrix -- rotate by angle about 'z' -- the "third" 
    ! dimension, i.e. out of the plane of the image
    R = as_rotmatrix(3, angle) 

    if (debug) print *,'rotation_matrix:',as_matrix(R)
    call rotate(R,vertices)

    if (debug) print *,'Post rotation'
    if (debug) print '(2F)',vertices(1:2,:)

    vertices(1,:) = vertices(1,:) + real(nx,rd_kind)/2.d0
    vertices(2,:) = vertices(2,:) + real(ny,rd_kind)/2.d0
    
    if (debug) print *,'Origin shifted back'
    if (debug) print '(2F)',vertices(1:2,:)

    if (present(missingvalue)) then
       if (present(coverage)) then
          call polySamp(data, vertices(1,:), vertices(2,:), output, missingvalue, coverage=coverage)
       else
          call polySamp(data, vertices(1,:), vertices(2,:), output, missingvalue)
       end if
    else
       call polySamp(data, vertices(1,:), vertices(2,:), output)
    end if

    ! call reorthogonalise(R)
    ! if (debug) print *,'rotation_matrix:',as_matrix(R)

  end subroutine rotate_image_real8_sub


  ! Subroutine interface


  ! In place rotation

  subroutine rotate_inplace(data, angle, missingvalue, coverage)

    ! Rotate and/or resize a single precision matrix, angle and missing value

    ! Interface variables
    real(kind=rd_kind), dimension(:,:), intent(inout) :: data
    real(kind=rd_kind), intent(in)                    :: angle
    real(kind=rd_kind), intent(in), optional          :: missingvalue
    real(kind=rs_kind), intent(in), optional          :: coverage

    if (present(missingvalue)) then
       if (present(coverage)) then
          call rotate_image(data,angle,data,missingvalue,coverage)
       else
          call rotate_image(data,angle,data,missingvalue)
       end if
    else
       call rotate_image(data,angle,data)
    end if

  end subroutine rotate_inplace

  subroutine rotate_inplace_integer_wrap(data, angle, missingvalue, coverage)

    ! Rotate and/or resize an integer matrix, single precision angle

    ! Interface variables
    integer, dimension(:,:), intent(inout)   :: data
    real(kind=rs_kind), intent(in)           :: angle
    integer, intent(in), optional            :: missingvalue
    real(kind=rs_kind), intent(in), optional :: coverage

    ! Local variables
    real(kind=rd_kind), dimension(size(data,1),size(data,2)) :: local_output

    if (present(missingvalue)) then
       if (present(coverage)) then
          call rotate_image(real(data,rd_kind),real(angle,rd_kind),local_output,real(missingvalue,rd_kind),coverage)
       else
          call rotate_image(real(data,rd_kind),real(angle,rd_kind),local_output,real(missingvalue,rd_kind))
       end if
    else
       call rotate_image(real(data,rd_kind),real(angle,rd_kind),local_output)
    end if

    data = nint(local_output)

  end subroutine rotate_inplace_integer_wrap

  subroutine rotate_inplace_integerd_wrap(data, angle, missingvalue, coverage)

    ! Rotate and/or resize an integer matrix, double precision angle

    ! Interface variables
    integer, dimension(:,:), intent(inout)   :: data
    real(kind=rd_kind), intent(in)           :: angle
    integer, intent(in), optional            :: missingvalue
    real(kind=rs_kind), intent(in), optional :: coverage

    ! Local variables
    real(kind=rd_kind), dimension(size(data,1),size(data,2)) :: local_output

    if (present(missingvalue)) then
       if (present(coverage)) then
          call rotate_image(real(data,rd_kind),angle,local_output,real(missingvalue,rd_kind),coverage)
       else
          call rotate_image(real(data,rd_kind),angle,local_output,real(missingvalue,rd_kind))
       end if
    else
       call rotate_image(real(data,rd_kind),angle,local_output)
    end if

    data = nint(local_output)

  end subroutine rotate_inplace_integerd_wrap

  subroutine rotate_inplace_real_wrap(data, angle, missingvalue, coverage)

    ! Rotate and/or resize a single precision matrix, angle and missing value

    ! Interface variables
    real(kind=rs_kind), dimension(:,:), intent(inout) :: data
    real(kind=rs_kind), intent(in)                    :: angle
    real(kind=rs_kind), intent(in), optional          :: missingvalue
    real(kind=rs_kind), intent(in), optional          :: coverage

    ! Local variables
    real(kind=rd_kind), dimension(size(data,1),size(data,2)) :: local_output

    if (present(missingvalue)) then
       if (present(coverage)) then
          call rotate_image(real(data,rd_kind),real(angle,rd_kind),local_output,real(missingvalue,rd_kind),coverage)
       else
          call rotate_image(real(data,rd_kind),real(angle,rd_kind),local_output,real(missingvalue,rd_kind))
       end if
    else
       call rotate_image(real(data,rd_kind),real(angle,rd_kind),local_output)
    end if

    data = real(local_output,rs_kind)

  end subroutine rotate_inplace_real_wrap

  subroutine rotate_inplace_reald_wrap(data, angle, missingvalue, coverage)

    ! Rotate and/or resize a single precision matrix, and missing value, double
    ! precision angle

    ! Interface variables
    real(kind=rs_kind), dimension(:,:), intent(inout) :: data
    real(kind=rd_kind), intent(in)                    :: angle
    real(kind=rs_kind), intent(in), optional          :: missingvalue
    real(kind=rs_kind), intent(in), optional          :: coverage

    ! Local variables
    real(kind=rd_kind), dimension(size(data,1),size(data,2)) :: local_output

    if (present(missingvalue)) then
       if (present(coverage)) then
          call rotate_image(real(data,rd_kind),angle,local_output,real(missingvalue,rd_kind),coverage)
       else
          call rotate_image(real(data,rd_kind),angle,local_output,real(missingvalue,rd_kind))
       end if
    else
       call rotate_image(real(data,rd_kind),angle,local_output)
    end if

    data = real(local_output,rs_kind)

  end subroutine rotate_inplace_reald_wrap


  ! Output to another variable

  subroutine rotate_image_integer_wrap(data, angle, output, missingvalue, coverage)

    ! Rotate and/or resize an integer matrix, single precision angle

    ! Interface variables
    integer, dimension(:,:), intent(in)      :: data
    real(kind=rs_kind), intent(in)           :: angle
    integer, dimension(:,:), intent(out)     :: output
    integer, intent(in), optional            :: missingvalue
    real(kind=rs_kind), intent(in), optional :: coverage

    ! Local variables
    real(kind=rd_kind), dimension(size(output,1),size(output,2)) :: local_output

    if (present(missingvalue)) then
       if (present(coverage)) then
          call rotate_image(real(data,rd_kind),real(angle,rd_kind),local_output,real(missingvalue,rd_kind),coverage)
       else
          call rotate_image(real(data,rd_kind),real(angle,rd_kind),local_output,real(missingvalue,rd_kind))
       end if
    else
       call rotate_image(real(data,rd_kind),real(angle,rd_kind),local_output)
    end if

    output = nint(local_output)

  end subroutine rotate_image_integer_wrap

  subroutine rotate_image_integerd_wrap(data, angle, output, missingvalue, coverage)

    ! Rotate and/or resize an integer matrix, double precision angle

    ! Interface variables
    integer, dimension(:,:), intent(in)      :: data
    real(kind=rd_kind), intent(in)           :: angle
    integer, dimension(:,:), intent(out)     :: output
    integer, intent(in), optional            :: missingvalue
    real(kind=rs_kind), intent(in), optional :: coverage

    ! Local variables
    real(kind=rd_kind), dimension(size(output,1),size(output,2)) :: local_output

    if (present(missingvalue)) then
       if (present(coverage)) then
          call rotate_image(real(data,rd_kind),angle,local_output,real(missingvalue,rd_kind))
       else
          call rotate_image(real(data,rd_kind),angle,local_output,real(missingvalue,rd_kind))
       end if
    else
       call rotate_image(real(data,rd_kind),angle,local_output)
    end if

    output = nint(local_output)

  end subroutine rotate_image_integerd_wrap

  subroutine rotate_image_real_wrap(data, angle, output, missingvalue, coverage)

    ! Rotate and/or resize a single precision matrix, angle and missing value

    ! Interface variables
    real(kind=rs_kind), dimension(:,:), intent(in)  :: data
    real(kind=rs_kind), intent(in)                  :: angle
    real(kind=rs_kind), dimension(:,:), intent(out) :: output
    real(kind=rs_kind), intent(in), optional        :: missingvalue
    real(kind=rs_kind), intent(in), optional        :: coverage

    ! Local variables
    real(kind=rd_kind), dimension(size(output,1),size(output,2)) :: local_output

    if (present(missingvalue)) then
       if (present(coverage)) then
          call rotate_image(real(data,rd_kind),real(angle,rd_kind),local_output,real(missingvalue,rd_kind),coverage)
       else
          call rotate_image(real(data,rd_kind),real(angle,rd_kind),local_output,real(missingvalue,rd_kind))
       end if
    else
       call rotate_image(real(data,rd_kind),real(angle,rd_kind),local_output)
    end if

    output = real(local_output,rs_kind)

  end subroutine rotate_image_real_wrap

  subroutine rotate_image_reald_wrap(data, angle, output, missingvalue, coverage)

    ! Rotate and/or resize a single precision matrix, angle and missing value

    ! Interface variables
    real(kind=rs_kind), dimension(:,:), intent(in)  :: data
    real(kind=rd_kind), intent(in)                  :: angle
    real(kind=rs_kind), dimension(:,:), intent(out) :: output
    real(kind=rs_kind), intent(in), optional        :: missingvalue
    real(kind=rs_kind), intent(in), optional        :: coverage

    ! Local variables
    real(kind=rd_kind), dimension(size(output,1),size(output,2)) :: local_output

    if (present(missingvalue)) then
       if (present(coverage)) then
          call rotate_image(real(data,rd_kind),angle,local_output,real(missingvalue,rd_kind),coverage)
       else
          call rotate_image(real(data,rd_kind),angle,local_output,real(missingvalue,rd_kind))
       end if
    else
       call rotate_image(real(data,rd_kind),angle,local_output)
    end if

    output = real(local_output,rs_kind)

  end subroutine rotate_image_reald_wrap
  

!!$  subroutine rotate_image_pnm_wrap(data, angle, output, missingvalue)
!!$
!!$    ! Rotate and/or resize a pnm object with single precision angle and missing value
!!$
!!$    ! Interface variables
!!$    type (pnm_object), intent(in)            :: data
!!$    real(kind=rs_kind), intent(in)           :: angle
!!$    type (pnm_object), intent(out)           :: output
!!$    real(kind=rs_kind), intent(in), optional :: missingvalue
!!$
!!$    ! Local variables
!!$    integer, dimension(size(pnm,1),size(pnm,2))                   :: local_input
!!$    real(kind=rd_kind), dimension(size(output,1),size(output,2)) :: local_output
!!$
!!$    local_input = data
!!$
!!$    if (present(missingvalue)) then
!!$       call rotate_image(real(local_input,rd_kind),real(angle,rd_kind),local_output,real(missingvalue,rd_kind))
!!$    else
!!$       call rotate_image(real(local_input,rd_kind),real(angle,rd_kind),local_output)
!!$    end if
!!$
!!$    output = nint(local_output)
!!$
!!$  end subroutine rotate_image_pnm_wrap
!!$
!!$  subroutine rotate_image_pnmd_wrap(data, angle, output, missingvalue)
!!$
!!$    ! Rotate and/or resize a pnm object with double precision angle and missing value
!!$
!!$    ! Interface variables
!!$    type (pnm_object), intent(in)            :: data
!!$    real(kind=rs_kind), intent(in)           :: angle
!!$    type (pnm_object), intent(out)           :: output
!!$    real(kind=rs_kind), intent(in), optional :: missingvalue
!!$
!!$    ! Local variables
!!$    integer, dimension(size(pnm,1),size(pnm,2))                   :: local_input
!!$    real(kind=rd_kind), dimension(size(output,1),size(output,2)) :: local_output
!!$
!!$    local_input = pnm
!!$
!!$    if (present(missingvalue)) then
!!$       call rotate_image(real(local_input,rd_kind),angle,local_output,missingvalue)
!!$    else
!!$       call rotate_image(real(local_input,rd_kind),angle,local_output)
!!$    end if
!!$
!!$    output = nint(local_output)
!!$
!!$  end subroutine rotate_image_pnmd_wrap


  ! Functional interface

  function rotate_resize_image_real8_fn(data, angle, newsize, missingvalue, coverage) result(rotimage)

    ! Rotate and resize a double precision image

    ! Interface variables
    real(kind=rd_kind), dimension(:,:), intent(in) :: data
    real(kind=rd_kind), intent(in)                 :: angle
    integer, dimension(2), intent(in)              :: newsize
    real(kind=rd_kind), intent(in), optional       :: missingvalue
    real(kind=rs_kind), intent(in), optional       :: coverage

    ! Function result
    real(kind=rd_kind), dimension(newsize(1),newsize(2)) :: rotimage

    if (present(missingvalue)) then
       if (present(coverage)) then
          call rotate_image(data,angle,rotimage,missingvalue,coverage)
       else
          call rotate_image(data,angle,rotimage,missingvalue)
       end if
    else
       call rotate_image(data,angle,rotimage)
    end if

  end function rotate_resize_image_real8_fn

  function rotate_image_real8_fn(data, angle, missingvalue, coverage) result(rotimage)

    ! Rotate a double precision image

    ! Interface variables
    real(kind=rd_kind), dimension(:,:), intent(in) :: data
    real(kind=rd_kind), intent(in)                 :: angle
    real(kind=rd_kind), intent(in), optional       :: missingvalue
    real(kind=rs_kind), intent(in), optional       :: coverage

    ! Function result
    real(kind=rd_kind), dimension(size(data,1),size(data,2)) :: rotimage

    if (present(missingvalue)) then
       if (present(coverage)) then
          call rotate_image(data,angle,rotimage,missingvalue,coverage)
       else
          call rotate_image(data,angle,rotimage,missingvalue)
       end if
    else
       call rotate_image(data,angle,rotimage)
    end if

  end function rotate_image_real8_fn

  function rotate_resize_image_real_fn(data, angle, newsize, missingvalue, coverage) result(rotimage)

    ! Rotate and resize a real image

    ! Interface variables
    real(kind=rs_kind), dimension(:,:), intent(in) :: data
    real(kind=rs_kind), intent(in)                 :: angle
    integer, dimension(2), intent(in)              :: newsize
    real(kind=rs_kind), intent(in), optional       :: missingvalue
    real(kind=rs_kind), intent(in), optional       :: coverage

    ! Function result
    real(kind=rs_kind), dimension(newsize(1),newsize(2)) :: rotimage

    if (present(missingvalue)) then
       if (present(coverage)) then
          call rotate_image(data,angle,rotimage,missingvalue,coverage)
       else
          call rotate_image(data,angle,rotimage,missingvalue)
       end if
    else
       call rotate_image(data,angle,rotimage)
    end if

  end function rotate_resize_image_real_fn

  function rotate_image_real_fn(data, angle, missingvalue, coverage) result(rotimage)

    ! Rotate a real image

    ! Interface variables
    real(kind=rs_kind), dimension(:,:), intent(in) :: data
    real(kind=rs_kind), intent(in)                 :: angle
    real(kind=rs_kind), intent(in), optional       :: missingvalue
    real(kind=rs_kind), intent(in), optional       :: coverage

    ! Function result
    real(kind=rs_kind), dimension(size(data,1),size(data,2)) :: rotimage

    if (present(missingvalue)) then
       if (present(coverage)) then
          call rotate_image(data,angle,rotimage,missingvalue,coverage)
       else
          call rotate_image(data,angle,rotimage,missingvalue)
       end if
    else
       call rotate_image(data,angle,rotimage)
    end if

  end function rotate_image_real_fn

  function rotate_resize_image_reald_fn(data, angle, newsize, missingvalue, coverage) result(rotimage)

    ! Rotate and resize a real image, double precision angle

    ! Interface variables
    real(kind=rs_kind), dimension(:,:), intent(in) :: data
    real(kind=rd_kind), intent(in)                 :: angle
    integer, dimension(2), intent(in)              :: newsize
    real(kind=rs_kind), intent(in), optional       :: missingvalue
    real(kind=rs_kind), intent(in), optional       :: coverage

    ! Function result
    real(kind=rs_kind), dimension(newsize(1),newsize(2)) :: rotimage

    if (present(missingvalue)) then
       if (present(coverage)) then
          call rotate_image(data,angle,rotimage,missingvalue,coverage)
       else
          call rotate_image(data,angle,rotimage,missingvalue)
       end if
    else
       call rotate_image(data,angle,rotimage)
    end if

  end function rotate_resize_image_reald_fn

  function rotate_image_reald_fn(data, angle, missingvalue, coverage) result(rotimage)

    ! Rotate a real image, double precision angle

    ! Interface variables
    real(kind=rs_kind), dimension(:,:), intent(in) :: data
    real(kind=rd_kind), intent(in)                 :: angle
    real(kind=rs_kind), intent(in), optional       :: missingvalue
    real(kind=rs_kind), intent(in), optional       :: coverage

    ! Function result
    real(kind=rs_kind), dimension(size(data,1),size(data,2)) :: rotimage

    if (present(missingvalue)) then
       if (present(coverage)) then
          call rotate_image(data,angle,rotimage,missingvalue,coverage)
       else
          call rotate_image(data,angle,rotimage,missingvalue)
       end if
    else
       call rotate_image(data,angle,rotimage)
    end if

  end function rotate_image_reald_fn

  function rotate_resize_image_integer_fn(data, angle, newsize, missingvalue, coverage) result(rotimage)

    ! Rotate and resize an image, single precision angle

    ! Interface variables
    integer, dimension(:,:), intent(in)      :: data
    real(kind=rs_kind), intent(in)           :: angle
    integer, dimension(2), intent(in)        :: newsize
    integer, intent(in), optional            :: missingvalue
    real(kind=rs_kind), intent(in), optional :: coverage

    ! Function result
    integer, dimension(newsize(1),newsize(2)) :: rotimage

    if (present(missingvalue)) then
       if (present(coverage)) then
          call rotate_image(data,angle,rotimage,missingvalue,coverage)
       else
          call rotate_image(data,angle,rotimage,missingvalue)
       end if
    else
       call rotate_image(data,angle,rotimage)
    end if

  end function rotate_resize_image_integer_fn

  function rotate_image_integer_fn(data, angle, missingvalue, coverage) result(rotimage)

    ! Rotate an image, single precision angle

    ! Interface variables
    integer, dimension(:,:), intent(in)      :: data
    real(kind=rs_kind), intent(in)           :: angle
    integer, intent(in), optional            :: missingvalue
    real(kind=rs_kind), intent(in), optional :: coverage

    ! Function result
    integer, dimension(size(data,1),size(data,2)) :: rotimage

    if (present(missingvalue)) then
       if (present(coverage)) then
          call rotate_image(data,angle,rotimage,missingvalue,coverage)
       else
          call rotate_image(data,angle,rotimage,missingvalue)
       end if
    else
       call rotate_image(data,angle,rotimage)
    end if

  end function rotate_image_integer_fn

  function rotate_resize_image_integerd_fn(data, angle, newsize, missingvalue, coverage) result(rotimage)

    ! Rotate and resize an image, double precision angle

    ! Interface variables
    integer, dimension(:,:), intent(in)      :: data
    real(kind=rd_kind), intent(in)           :: angle
    integer, dimension(2), intent(in)        :: newsize
    integer, intent(in), optional            :: missingvalue
    real(kind=rs_kind), intent(in), optional :: coverage

    ! Function result
    integer, dimension(newsize(1),newsize(2)) :: rotimage

    if (present(missingvalue)) then
       if (present(coverage)) then
          call rotate_image(data,angle,rotimage,missingvalue,coverage)
       else
          call rotate_image(data,angle,rotimage,missingvalue)
       end if
    else
       call rotate_image(data,angle,rotimage)
    end if

  end function rotate_resize_image_integerd_fn

  function rotate_image_integerd_fn(data, angle, missingvalue, coverage) result(rotimage)

    ! Rotate an image, double precision angle

    ! Interface variables
    integer, dimension(:,:), intent(in)      :: data
    real(kind=rd_kind), intent(in)           :: angle
    integer, intent(in), optional            :: missingvalue
    real(kind=rs_kind), intent(in), optional :: coverage

    ! Function result
    integer, dimension(size(data,1),size(data,2)) :: rotimage

    if (present(missingvalue)) then
       if (present(coverage)) then
          call rotate_image(data,angle,rotimage,missingvalue,coverage)
       else
          call rotate_image(data,angle,rotimage,missingvalue)
       end if
    else
       call rotate_image(data,angle,rotimage)
    end if

  end function rotate_image_integerd_fn

!!$  function rotate_resize_image_pnm_fn(data, angle, newsize, missingvalue) result(rotimage)
!!$
!!$    ! Rotate and resize a pnm image
!!$
!!$    ! Interface variables
!!$    type (pnm_object), intent(in)     :: data
!!$    real(kind=rs_kind), intent(in)               :: angle
!!$    integer, dimension(2), intent(in) :: newsize
!!$    integer, intent(in), optional     :: missingvalue
!!$
!!$    ! Function result
!!$    type (pnm_object) :: rotimage
!!$
!!$    ! Local variable
!!$    integer, dimension(newsize(1),newsize(2)) :: local_output
!!$
!!$    if (present(missingvalue)) then
!!$       call rotate_image(as_array_2d(data),angle,local_output,missingvalue)
!!$    else
!!$       call rotate_image(as_array_2d(data),angle,local_output)
!!$    end if
!!$
!!$    rotimage = local_output
!!$
!!$  end function rotate_resize_image_pnm_fn
!!$
!!$  function rotate_image_pnm_fn(data, angle, missingvalue) result(rotimage)
!!$
!!$    ! Rotate an image
!!$
!!$    ! Interface variables
!!$    type(pnm_object), intent(in)   :: data
!!$    real(kind=rs_kind), intent(in) :: angle
!!$    integer, intent(in), optional  :: missingvalue
!!$
!!$    ! Function result
!!$    type(pnm_object) :: rotimage
!!$
!!$    if (present(missingvalue)) then
!!$       call rotate_image(data,angle,rotimage,missingvalue)
!!$    else
!!$       call rotate_image(data,angle,rotimage)
!!$    end if
!!$
!!$  end function rotate_image_pnm_fn
  
  subroutine rotave_real8_sub(data, nfold, missingvalue, coverage)

    ! Rotationally average in-place a double precision matrix with 
    ! missing value defined.

    ! Interface variables
    real(kind=rd_kind), dimension(:,:), intent(inout) :: data
    integer, intent(in)                               :: nfold
    real(kind=rd_kind), intent(in)                    :: missingvalue
    real(kind=rs_kind), intent(in), optional          :: coverage

    ! Local variables
    real(rd_kind), dimension(size(data,1),size(data,2)) :: buffer, totbuffer
    integer, dimension(size(data,1),size(data,2))       :: bufcount, totcount
    real(kind=rd_kind) :: step
    integer :: i, j, k

    ! Initialise the 'total' buffer with the original data
    totbuffer = data

    ! Ditto for the 'total' counts buffer (ignoring missingvalues)
    totcount = 0
    where(totbuffer /= missingvalue) totcount = 1
    
    ! Precompute the rotation corresponding to an nfold rotation
    step = (2.d0/real(nfold,rd_kind))*pi
    
    ! Now step through 
    do i = 1, nfold-1
       buffer = data
       if (present(coverage)) then
          call rotate_image(buffer,real(i,rd_kind)*step,missingvalue,coverage)
       else
          call rotate_image(buffer,real(i,rd_kind)*step,missingvalue)
       end if
       do j = 1, size(data,1)
          do k = 1, size(data,2)
             if (buffer(j,k) /= missingvalue) then
                if (totbuffer(j,k) /= missingvalue) then
                   totbuffer(j,k) = totbuffer(j,k) + buffer(j,k)
                else
                   totbuffer(j,k) = buffer(j,k)
                end if
                totcount(j,k) = totcount(j,k) + 1
             end if
          end do
       end do
    end do

    where(totcount > 0) totbuffer = totbuffer/real(totcount)

    data = totbuffer

  end subroutine rotave_real8_sub
  
  subroutine rotave_resize_real8_sub(data, nfold, output, missingvalue, coverage)

    ! Rotationally average a double precision matrix with missing 
    ! value defined. If output is a different size to data, this 
    ! will be done with re-sizing.

    ! Interface variables
    real(kind=rd_kind), dimension(:,:), intent(in)  :: data
    integer, intent(in)                             :: nfold
    real(kind=rd_kind), dimension(:,:), intent(out) :: output
    real(kind=rd_kind), intent(in)                  :: missingvalue
    real(kind=rs_kind), intent(in), optional        :: coverage

    ! Local variables
    real(rd_kind), dimension(size(data,1),size(data,2))     :: inbuffer
    real(rd_kind), dimension(size(output,1),size(output,2)) :: outbuffer, totbuffer
    integer, dimension(size(output,1),size(output,2))       :: bufcount, totcount
    real(kind=rd_kind) :: step
    integer :: i, j, k

    totcount = 0
    totbuffer = missingvalue
    
    step = (2.d0/real(nfold,rd_kind))*pi
    
    do i = 0, nfold-1
       inbuffer = data
       outbuffer = missingvalue
       if (present(coverage)) then
          call rotate_image(inbuffer,real(i,rd_kind)*step,outbuffer,missingvalue,coverage)
       else
          call rotate_image(inbuffer,real(i,rd_kind)*step,outbuffer,missingvalue)
       end if
       do j = 1, size(output,1)
          do k = 1, size(output,2)
             if (outbuffer(j,k) /= missingvalue) then
                if (totbuffer(j,k) /= missingvalue) then
                   totbuffer(j,k) = totbuffer(j,k) + outbuffer(j,k)
                else
                   totbuffer(j,k) = outbuffer(j,k)
                end if
                totcount(j,k) = totcount(j,k) + 1
             end if
          end do
       end do
    end do

    where(totcount > 0) totbuffer = totbuffer/real(totcount)

    output = totbuffer  

  end subroutine rotave_resize_real8_sub

  subroutine rotave_real8_sub_nomiss(data, nfold)

    ! Rotationally average a double precision matrix with no missing 
    ! value defined. If output is a different size to data, this 
    ! will be done with re-sizing.

    ! Interface variables
    real(kind=rd_kind), dimension(:,:), intent(inout) :: data
    integer, intent(in)                               :: nfold

    ! Local variables
    real(rd_kind), dimension(size(data,1),size(data,2)) :: buffer, tot
    real(kind=rd_kind) :: step
    integer :: i

    ! Precompute the rotation to create an n-fold average
    step = (2.d0/real(nfold,rd_kind))*pi

    ! Initialise our output to the same as in the input
    tot = data

    ! Now add nfold minus 1 rotated copies on top ...
    do i = 1, nfold-1
       buffer = data
       call rotate_image(buffer,real(i,rd_kind)*step)
       tot = tot + buffer
    end do

    ! ... and divide through by nfold
    data = tot/real(nfold,kind=rd_kind)

  end subroutine rotave_real8_sub_nomiss

  subroutine rotave_resize_real8_sub_nomiss(data, nfold, output)

    ! Rotationally average a double precision matrix with no missing 
    ! value defined.

    ! Interface variables
    real(kind=rd_kind), dimension(:,:), intent(in)  :: data
    integer, intent(in)                             :: nfold
    real(kind=rd_kind), dimension(:,:), intent(out) :: output

    ! Local variables
    real(rd_kind), dimension(size(data,1),size(data,2)) :: inbuffer
    real(rd_kind), dimension(size(output,1),size(output,2)) :: outbuffer, tot
    real(kind=rd_kind) :: step
    integer :: i

    ! Precompute the rotation to create an n-fold average
    step = (2.d0/real(nfold,rd_kind))*pi

    ! Initialise our output to zero this time
    tot = 0.d0
    
    ! And then create nfold versions of the original, possibly
    ! with resizing ...
    do i = 0, nfold-1
       inbuffer = data
       call rotate_image(inbuffer,real(i,rd_kind)*step,outbuffer)
       tot = tot + outbuffer
    end do

    output = tot/real(nfold,kind=rd_kind)

  end subroutine rotave_resize_real8_sub_nomiss

!!$  subroutine rotave_real8_inplace(data, nfold, missingvalue, coverage)
!!$
!!$    ! Wrapper to allow in-place rotational averaging
!!$
!!$    ! Interface variables
!!$    real(kind=rd_kind), dimension(:,:), intent(inout) :: data
!!$    integer, intent(in)                               :: nfold
!!$    real(kind=rd_kind), intent(in), optional          :: missingvalue
!!$    real(kind=rs_kind), intent(in), optional          :: coverage
!!$
!!$    ! Local variables
!!$    real(kind=rd_kind), dimension(size(data,1),size(data,2)) :: output
!!$
!!$    if (present(missingvalue)) then
!!$       if (present(coverage)) then
!!$          call rotave_image(data, nfold, output, missingvalue, coverage)
!!$       else
!!$          call rotave_image(data, nfold, output, missingvalue)
!!$       end if
!!$    else
!!$       call rotave_image(data, nfold, output)
!!$    end if
!!$
!!$    data = output
!!$
!!$  end subroutine rotave_real8_inplace

  subroutine rotave_real_sub(data, nfold, output, missingvalue, coverage)

    ! Interface variables
    real(kind=rs_kind), dimension(:,:), intent(in)  :: data
    integer, intent(in)                             :: nfold
    real(kind=rs_kind), dimension(:,:), intent(out) :: output
    real(kind=rs_kind), intent(in), optional        :: missingvalue
    real(kind=rs_kind), intent(in), optional        :: coverage

    ! Local variables
    real(kind=rd_kind), dimension(size(output,1),size(output,2)) :: localoutput

    if (present(missingvalue)) then
       if (present(coverage)) then
          call rotave_image(real(data,rd_kind), nfold, localoutput, real(missingvalue,rd_kind), coverage)
       else
          call rotave_image(real(data,rd_kind), nfold, localoutput, real(missingvalue,rd_kind))
       end if
    else
       call rotave_image(real(data,rd_kind), nfold, localoutput)
    end if

    output = real(localoutput,rs_kind)

  end subroutine rotave_real_sub

  subroutine rotave_integer_sub(data, nfold, output, missingvalue, coverage)

    ! Interface variables
    integer, dimension(:,:), intent(in)      :: data
    integer, intent(in)                      :: nfold
    integer, dimension(:,:), intent(out)     :: output
    integer, intent(in), optional            :: missingvalue
    real(kind=rs_kind), intent(in), optional :: coverage

    ! Local variables
    real(kind=rd_kind), dimension(size(output,1),size(output,2)) :: localoutput

    if (present(missingvalue)) then
       if (present(coverage)) then
          call rotave_image(real(data,rd_kind), nfold, localoutput, real(missingvalue,rd_kind), coverage)
       else
          call rotave_image(real(data,rd_kind), nfold, localoutput, real(missingvalue,rd_kind))
       end if
    else
       call rotave_image(real(data,rd_kind), nfold, localoutput)
    end if

    output = nint(localoutput,rs_kind)

  end subroutine rotave_integer_sub

  subroutine rotave_real_inplace_sub(data, nfold, missingvalue, coverage)

    ! Interface variables
    real(kind=rs_kind), dimension(:,:), intent(inout) :: data
    integer, intent(in)                               :: nfold
    real(kind=rs_kind), intent(in), optional          :: missingvalue
    real(kind=rs_kind), intent(in), optional          :: coverage

    ! Local variables
    real(kind=rd_kind), dimension(size(data,1),size(data,2)) :: localoutput

    if (present(missingvalue)) then
       if (present(coverage)) then
          call rotave_image(real(data,rd_kind), nfold, localoutput, real(missingvalue,rd_kind), coverage)
       else
          call rotave_image(real(data,rd_kind), nfold, localoutput, real(missingvalue,rd_kind))
       end if
    else
       call rotave_image(real(data,rd_kind), nfold, localoutput)
    end if

    data = real(localoutput,rs_kind)

  end subroutine rotave_real_inplace_sub

  subroutine rotave_integer_inplace_sub(data, nfold, missingvalue, coverage)

    ! Interface variables
    integer, dimension(:,:), intent(inout)   :: data
    integer, intent(in)                      :: nfold
    integer, intent(in), optional            :: missingvalue
    real(kind=rs_kind), intent(in), optional :: coverage

    ! Local variables
    real(kind=rd_kind), dimension(size(data,1),size(data,2)) :: localoutput

    if (present(missingvalue)) then
       if (present(coverage)) then
          call rotave_image(real(data,rd_kind), nfold, localoutput, real(missingvalue,rd_kind), coverage)
       else
          call rotave_image(real(data,rd_kind), nfold, localoutput, real(missingvalue,rd_kind))
       end if
    else
       call rotave_image(real(data,rd_kind), nfold, localoutput)
    end if

    data = nint(localoutput,rs_kind)

  end subroutine rotave_integer_inplace_sub

  function rotave_real8_fn(data, nfold, missingvalue, coverage) result(rotimage)

    ! Rotate a real image, double precision angle

    ! Interface variables
    real(kind=rd_kind), dimension(:,:), intent(in) :: data
    integer, intent(in)                            :: nfold
    real(kind=rd_kind), intent(in), optional       :: missingvalue
    real(kind=rs_kind), intent(in), optional       :: coverage

    ! Function result
    real(kind=rd_kind), dimension(size(data,1),size(data,2)) :: rotimage

    if (present(missingvalue)) then
       if (present(coverage)) then
          call rotave_image(data,nfold,rotimage,missingvalue,coverage)
       else
          call rotave_image(data,nfold,rotimage,missingvalue)
       end if
    else
       call rotave_image(data,nfold,rotimage)
    end if

  end function rotave_real8_fn

  function rotave_real_fn(data, nfold, missingvalue, coverage) result(rotimage)

    ! Rotate a real image, double precision angle

    ! Interface variables
    real(kind=rs_kind), dimension(:,:), intent(in) :: data
    integer, intent(in)                            :: nfold
    real(kind=rs_kind), intent(in), optional       :: missingvalue
    real(kind=rs_kind), intent(in), optional       :: coverage

    ! Function result
    real(kind=rs_kind), dimension(size(data,1),size(data,2)) :: rotimage

    if (present(missingvalue)) then
       if (present(coverage)) then
          call rotave_image(data,nfold,rotimage,missingvalue,coverage)
       else
          call rotave_image(data,nfold,rotimage,missingvalue)
       end if
    else
       call rotave_image(data,nfold,rotimage)
    end if

  end function rotave_real_fn
  
  function rotave_integer_fn(data, nfold, missingvalue, coverage) result(rotimage)

    ! Rotate a real image, double precision angle

    ! Interface variables
    integer, dimension(:,:), intent(in)      :: data
    integer, intent(in)                      :: nfold
    integer, intent(in), optional            :: missingvalue
    real(kind=rs_kind), intent(in), optional :: coverage

    ! Function result
    integer, dimension(size(data,1),size(data,2)) :: rotimage

    if (present(missingvalue)) then
       if (present(coverage)) then
          call rotave_image(data,nfold,rotimage,missingvalue,coverage)
       else
          call rotave_image(data,nfold,rotimage,missingvalue)
       end if
    else
       call rotave_image(data,nfold,rotimage)
    end if

  end function rotave_integer_fn

  subroutine centre_real8_inplace (data, missingvalue)

    real(kind=rd_kind), dimension(:,:), intent(inout) :: data
    real(kind=rd_kind), intent(in), optional          :: missingvalue

    ! Local variables
    logical, dimension(size(data))            :: notmissingmask
    real(kind=rd_kind), dimension(size(data)) :: oneddata 
    integer, dimension(size(data))            :: indices
    real(kind=rd_kind)                        :: xcentre, ycentre
    integer, dimension(:), pointer            :: clustertmp

    integer :: i, npixels, nx, ny

    nx = size(data,1)
    ny = size(data,2)

    oneddata = pack(data, .TRUE.)

    if (present(missingvalue)) then
       notmissingmask = (oneddata /= missingvalue)
    else
       notmissingmask = .TRUE.
    end if
    
    indices = (/ (i, i=1,size(indices)) /)

    npixels = push(clustertmp, pack(indices, mask=notmissingmask))

    xcentre = sum(mod(clustertmp,nx))/npixels
    ycentre = sum((clustertmp/ny)+1)/npixels

    ! Remove all the elements from the clustertmp array
    npixels = splice(clustertmp, 0)

    print *,xcentre,ycentre

  end subroutine centre_real8_inplace

end module image_transforms
